<?php

// ci.php
// Jeff Schwentner
// Support for ci.php...

require_once HORDE_BASE . '/lib/Perms.php';

/**
* Checks in an RCS file with a specified change log.
*
* @param filepath  Location of file to checkin.
* @param message   Log of changes since last version.
* @param new_rev   Returns the new revision number, if successful.
*/

function check_in($filepath, $message = 'No comment specified', $newBinary = false, $new_rev = '')
{
    if(!isset($GLOBALS['conf']['paths']['ci'])){
        Chora::fatal("Cannot find path to ci.exe");
    }
    
    // Get username for rcs tag...
    $user_name = Auth::getAuth();

    if($user_name) putenv("LOGNAME=$user_name");
    else putenv("LOGNAME=guest");

    $Q = 'OS_WINDOWS' ? '"' : "'" ;
    
    $ci_cmd  = $GLOBALS['conf']['paths']['ci'] . ' ' . $Q . $filepath . $Q.' 2>&1';
    $rcs_cmd = $GLOBALS['conf']['paths']['rcs'] . ' -i -kb ' . $Q . $filepath . $Q.' 2>&1';
    $output = "";

    $message_lines = explode("\n", $message);

    $pipe_def = array(
       0 => array("pipe", "r"),
       1 => array("pipe", "w"));

    if($newBinary) {
        $process = proc_open($rcs_cmd, $pipe_def, $pipes);
    } 
    else {
        $process = proc_open($ci_cmd, $pipe_def, $pipes);
    }
    
    if (is_resource($process)) {
        foreach ($message_lines as $line)
        {           
            if($line == '.\n') $line = '. \n';
            fwrite($pipes[0], $line);
        }
       
        fwrite($pipes[0], "\n.\n");
        fclose($pipes[0]);
        
        while(!feof($pipes[1]))
        {
            $output .= fread($pipes[1], 8192);
        }
        fclose($pipes[1]);
        proc_close($process);
    }
    else 
    {
        Chora::fatal("Failed to open pipe in check_in(...).");
    }
    
    if($newBinary)
    {
        exec($ci_cmd.' 2>&1', $return_array, $retval);

        if ($retval) 
        {
            Chora::fatal("Unable to spawn ci on $filepath from checkin(...).");
        } else {
            foreach ($return_array as $line) {
                $output .= $line;
            }
        }
    }

    $rev_start = strpos($output, 'new revision: ');
    
    // If no new revision, see if this is an initial checkin...
    if($rev_start === false)
    {
        $rev_start = strpos($output, 'initial revision: ');
        $rev_end   = strpos($output, ' ', $rev_start);
    }
    else
    {
        $rev_end   = strpos($output, ';', $rev_start);
    }
    
    if($rev_start !== false && $rev_end !== false)
    {
        $rev_start += 14;
        $new_rev = substr($output, $rev_start, $rev_end - $rev_start);
        return true;
    }
    else 
    {
        //unlock_file($filepath);
        $temp_pos = strpos($output, 'file is unchanged');
        if($temp_pos !== false)
        {
            Chora::fatal("Check-in Failure: ".basename($filepath)." has not been modified.");
        }
        else
        {
            Chora::fatal("Failed to checkin $filepath, $ci_cmd, $output.");
        }
    }
}


/**
* Checks the locks on an CVS/RCS file.
*
* @param  filepath  Location of file.
* @param  locked_by Returns the username holding the lock.
* @return boolean   True if file is locked, false otherwise.
*/

function file_locked($filepath, $locked_by = '')
{
    $rlog_cmd  = $GLOBALS['conf']['paths']['rlog'];
    $rlog_flag = " -L ";

    $Q = 'OS_WINDOWS' ? '"' : "'" ;

    $cmd = $rlog_cmd . $rlog_flag . $Q . $filepath . $Q;

    exec($cmd.' 2>&1', $return_array, $retval);

    if ($retval) 
    {
        Chora::fatal("Unable to spawn rlog on $filepath from file_locked(...).");
    }
    else
    {
        $output = '';
    
        foreach ($return_array as $line)
        {
            $output .= $line;
        }

        $start_name = strpos($output, 'locked by: ');
        $end_name   = strpos($output, ';', $start_name);
    
        if($start_name !== false && $end_name !== false)
        {
            $start_name += 11;
            $locked_by = substr($output, $start_name, $end_name - $start_name);
            return true;
        }
        else if(strlen($output) == 0)
        {
            return false;
        }
        else
        {
            Chora::fatal("$Failure running rlog in file_locked(...).");
        }
    }
}


/**
* Locks an CVS/RCS file.
*
* @param filepath  Location of file.
*/

function lock_file($filepath)
{

    // Get username for lock rcs tag...
    $user_name = Auth::getAuth();

    if($user_name) putenv("LOGNAME=$user_name");
    else putenv("LOGNAME=guest");

    $rcs_cmd  = $GLOBALS['conf']['paths']['rcs'];
    $rcs_flag = " -l ";

    $Q = 'OS_WINDOWS' ? '"' : "'" ;

    $cmd = $rcs_cmd . $rcs_flag . $Q . $filepath . $Q;

    exec($cmd.' 2>&1', $return_array, $retval);

    if ($retval) 
    {
        Chora::fatal("Failed to spawn rcs on $filepath from lock_file(...).");
    }
    else
    {
        $output = '';
    
        foreach ($return_array as $line)
        {
            $output .= $line;
        }
        
        $locked_pos = strpos($output, 'locked');

        if($locked_pos !== false)
        {
            return true;
        }
        else
        {
            Chora::fatal("lock_file(..) ($retval) $cmd $output.");

            return false;
        }
    }
}


/**
* Unlocks an CVS/RCS file.
*
* @param filepath  Location of file.
*/

function unlock_file($filepath)
{
    // Get username for rcs tag...
    $user_name = Auth::getAuth();

    if($user_name) putenv("LOGNAME=$user_name");
    else putenv("LOGNAME=guest");
    
    $rcs_cmd  = $GLOBALS['conf']['paths']['rcs'];
    $rcs_flag = " -u ";

    $Q = 'OS_WINDOWS' ? '"' : "'" ;

    $cmd = $rcs_cmd . $rcs_flag . $Q . $filepath . $Q;

    exec($cmd.' 2>&1', $return_array, $retval);

    if ($retval) 
    {
        Chora::fatal("Failed to spawn rcs on $filepath from unlock_file(...). $cmd");
    }
    else
    {
        $output = '';
    
        foreach ($return_array as $line)
        {
            $output .= $line;
        }
        
        $unlocked_pos = strpos($output, 'unlocked');

        if($unlocked_pos !== false)
        {
            return true;
        }
        else
        {
            // Already unlocked...
            return true;
        }
    }
}

/**
* Retrieves the user permissions for a given repository location.
*
* @param  sourceName    Name of the repository
* @param  permLocation  Location of the file or directory
* @param  userName      User to fetch permissions for
* @return integer       Numerical permission value
*/

function getLocationPerms($sourceName, $permLocation = '', $userName = '')
{
    global $conf;
    if(!$userName)
    {
        $userName = Auth::getAuth();
    }

    if(!$permLocation)
    {
        $permLocation = $sourceName;
    }

    else
    {
        $permLocation = $sourceName.'/'.$permLocation;
        $permLocation = preg_replace('/(\\/)(\\/)/', '/', $permLocation);
    }

    $permLocation = preg_replace('/(\\/)(\\/)/', '/', $permLocation);

    $permsContainer = &new Perms();
    $permsArray = explode('/', $permLocation);

    while($permsArray)
    {
        $permKey = implode('/', $permsArray);
        if($permsContainer->exists($permKey))
        {
        
            $permsChild = &$permsContainer->getPermission($permKey);
            $userPerms  = $permsChild->getUserPermissions();
            if(isset($userPerms[$userName]))
            {
                return $userPerms[$userName];
            }
            else 
            {
                return false;
            }
        }
        array_pop($permsArray);
    }
    return false;
}

/**
* Sets the user permissions for a given repository location.
*
* @param perms         Permissions to apply for user
* @param sourceName    Name of the repository
* @param permLocation  Location of the file or directory
* @param userName      User to set permissions for
*/

function addLocationPerms($perms, $sourceName, $permLocation = '', $userName = '')
{
    global $conf;
    if(!$userName)
    {
        $userName = Auth::getAuth();
    }

    if(!$permLocation)
    {
        $permLocation = $sourceName;
    }

    else
    {
        $permLocation = $sourceName.'/'.$permLocation;
        $permLocation = preg_replace('/(\\/)(\\/)/', '/', $permLocation);
    }

    $permsContainer = &new Perms();

    if($permsContainer->exists($permLocation))
    {
        $permsChild = &$permsContainer->getPermission($permLocation);
    }
    
    else
    {
        $permsChild = &$permsContainer->newPermission($permLocation);
    }
    
    $permsChild->addUserPermission($userName, $perms, false);
    $permsContainer->addPermission($permsChild);
    $permsChild->save();
}


?>