<?php
/**
 * VC_cvs implementation.
 *
 * Copyright 2000-2003 Anil Madhavapeddy, <anil@recoil.org>
 *
 * $Horde: framework/VC/VC/cvs.php,v 1.3 2003/10/30 23:38:18 jan Exp $
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @version $Revision: 1.3 $
 * @since   Chora 0.1
 * @package VC
 */
class VC_cvs extends VC {

    var $cvsusers = null;

    /**
     * Return the CVSROOT for this repository, with no trailing /
     *
     * @return string  CVSROOT for this repository.
     */
    function sourceroot()
    {
        return $GLOBALS['conf']['paths']['sourceroot'];
    }

    function &isFile($where)
    {
        return @is_file($where . ',v');
    }

    function &queryDir($where)
    {
        return $ret = &new CVSLib_Directory($this, $where);
    }

    function getCheckout($file, $rev)
    {
        return CVSLib_Checkout::get($file->queryFullPath(), $rev);
    }

    function &getDiff(&$file, $rev1, $rev2, $type = 'context', $num = 3, $ws = true)
    {
        return CVSLib_Diff::get($file, $rev1, $rev2, $type, $num, $ws);
    }

    function &getFileObject($filename, $quicklog = false)
    {
        return CVSLib_File::getFileObject($this, $this->sourceroot() . '/' . $filename, $quicklog);
    }

    function &getAnnotateObject($filename)
    {
        return $ret = &new CVSLib_Annotate($this, $filename);
    }

    function &getPatchsetObject($filename)
    {
        return CVSLib_Patchset::getPatchsetObject($this, $this->sourceroot() . '/' . $filename);
    }

    /**
     * Parse the 'cvsusers' file, if present in the CVSROOT, and
     * return a hash containing the requisite information, keyed on
     * the username, and with the 'desc','name', and 'mail' values
     * inside.  @return false if the file is not present, otherwise
     * populate $this->cvsusers with the data
     */
    function parseCVSUsers()
    {
        /* Check that we haven't already parsed cvsusers. */
        if (is_array($this->cvsusers)) {
            return true;
        }

        /* Try to locate the cvsusers file, and test to see if it is
         * there. */
        $cvsfile = $GLOBALS['conf']['paths']['cvsusers'];

        if (!@is_file($cvsfile) || !($fl = @fopen($cvsfile, OS_WINDOWS ? 'rb' : 'r'))) {
            return false;
        }

        $this->cvsusers = array();

        /* Discard the first line, since it'll be the header info. */
        fgets($fl, 4096);

        /* Parse the rest of the lines into a hash, keyed on
         * username. */
        while ($line = fgets($fl, 4096)) {
            if (preg_match('/^\s*$/', $line)) continue;
            if (!preg_match('/^(\w+)\s+(.+)\s+([\w\.\-\_]+@[\w\.\-\_]+)\s+(.*)$/', $line, $regs)) continue;
            $this->cvsusers[$regs[1]]['name'] = trim($regs[2]);
            $this->cvsusers[$regs[1]]['mail'] = trim($regs[3]);
            $this->cvsusers[$regs[1]]['desc'] = trim($regs[4]);
        }

        return true;
    }

}

/**
 * CVSLib annotate class.
 *
 * See the README file that came with this library for more
 * information, and read the inline documentation.
 *
 * Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @version $Revision: 1.3 $
 * @since   Chora 0.2
 * @package VC
 */
class CVSLib_Annotate {

    var $file;
    var $rep;
    var $tmpfile;

    function CVSLib_Annotate(&$rep, $file)
    {
        $this->rep = &$rep;
        $this->file = &$file;
    }

    function doAnnotate($rev)
    {
        global $conf;

        /* Make sure that the file values for this object is valid */
        if (is_a($this->file, 'PEAR_Error')) {
            return false;
        }

        /* Make sure that the cvsrep parameter is valid */
        if (is_a($this->rep, 'PEAR_Error')) {
            return false;
        }

        if (!VC_Revision::valid($rev)) {
            return false;
        }

        $where = $this->file->queryModulePath();
        $sourceroot = $this->rep->sourceroot();

        $this->tmpfile = Util::getTempFile('chora');

        $pipe = popen($conf['paths']['cvs'] . ' -n -l server > ' . $this->tmpfile, OS_WINDOWS ? 'wb' : 'w');

        $out = array();
        $out[] = "Root $sourceroot";
        $out[] = 'Valid-responses ok error Valid-requests Checked-in Updated Merged Removed M E';
        $out[] = 'UseUnchanged';
        $out[] = 'Argument -r';
        $out[] = "Argument $rev";
        $out[] = "Argument $where";
        $dirs = explode('/', dirname($where));
        while (sizeof($dirs)) {
            $out[] = 'Directory ' . implode('/', $dirs);
            $out[] = "$sourceroot/" . implode('/', $dirs);
            array_pop($dirs);
        }
        $out[] = 'Directory .';
        $out[] = $sourceroot;
        $out[] = 'annotate';

        foreach ($out as $line) {
            fwrite($pipe, "$line\n");
        }

        pclose($pipe);

        if (!($fl = fopen($this->tmpfile, OS_WINDOWS ? 'rb' : 'r'))) {
            exit;
        }

        $lines = array();
        $line = fgets($fl, 4096);

        // Windows versions of cvs always return $where with forwards slashes.
        if (OS_WINDOWS) {
            $where = str_replace(DIRECTORY_SEPARATOR, '/', $where);
        }    

        while ($line && !preg_match("|^E\s+Annotations for $where|", $line)) {
            $line = fgets($fl, 4096);
        }

        if (!$line) {
            return PEAR::raiseError(sprintf(_("Unable to annotate; server said: %s"), $line));
        }

        $lineno = 1;
        while ($line = fgets($fl, 4096)) {
            if (preg_match('/^M\s+([\d\.]+)\s+\((\w+)\s+(\d+-\w+-\d+)\):.(.*)$/', $line, $regs)) {
                $entry = array();
                $entry['rev']    = $regs[1];
                $entry['author'] = $regs[2];
                $entry['date']   = $regs[3];
                $entry['line']   = $regs[4];
                $entry['lineno'] = $lineno++;
                $lines[] = $entry;
            }
        }

        fclose($fl);
        return $lines;
    }

}

/**
 * CVSLib checkout class.
 *
 * See the README file that came with this library for more
 * information, and read the inline documentation.
 *
 * Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @version $Revision: 1.3 $
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_Checkout {

    /**
      * Static function which returns a file pointing to the head of the requested
      * revision of an RCS file.
      * @param fullname Fully qualified pathname of the desired RCS file to checkout
      * @param rev RCS revision number to check out
      * @return Either a PEAR_Error object, or a stream pointer to the head of the checkout
      */
    function get($fullname, $rev)
    {
        if (!VC_Revision::valid($rev)) {
            return PEAR::raiseError(_("Invalid revision number"));
        }

        if (OS_WINDOWS) {
            $Q = '"';
            $mode = 'rb';
        } else {
            $Q = "'";
            $mode = 'r';
        }

        if (!($RCS = popen($GLOBALS['conf']['paths']['co'] . " -p$rev $Q$fullname$Q 2>&1", $mode))) {
            return PEAR::raiseError(_("Couldn't perform checkout of the requested file"));
        }

        /* First line from co should be of the form :
         * /path/to/filename,v  -->  standard out
         * and we check that this is the case and error otherwise
         */

        $co = fgets($RCS, 1024);
        
        if (!preg_match('/^([\S ]+),v\s+-->\s+st(andar)?d ?out(put)?\s*$/', $co, $regs) || $regs[1].',v' != $fullname) {
            return PEAR::raiseError(sprintf(_("Unexpected output from Checkout: %s"), $co));
        }

        /* Next line from co is of the form:
         * revision 1.2.3
         * TODO: compare this to $rev for consistency, atm we just
         *       discard the value to move input pointer along - avsm
         */
        $co = fgets($RCS, 1024);

        return $RCS;
    }

    /**
    * Compresses the files and folders in the "$where" directory
    * into a zipfile object.
    *
    * @param where    Location of the target directory
    * @param zipfile  zipfile object
    * @param recurse  Specifies whether or not to recurse (true)
    */

    function zip_dir($where, $zipfile, $recurse = true)
    {
        global $VC;
        global $conf;

        static $recurse_count = 0;
        static $repository_name = "";
        static $zip_log = "";
    
        if($recurse_count == 0)
        {
            $date_time = getdate(); 
        
            $date_str =  $date_time['mon'].'/'.
                         $date_time['mday'].'/'.
                         $date_time['year'];

            $time_str  = $date_time['hours'].':'.
                         $date_time['minutes'].':'.
                         $date_time['seconds'];

            $repository_name = $conf['options']['sourceRootName'];
        
            $zip_log = "Repository: $repository_name\r\n".
                       "Date: $date_str\r\n".
                       "Time: $time_str\r\n\r\n".
                       "Ver\tFile\r\n".
                       "---\t----\r\n";
        }

        $recurse_count++;

        Chora::checkError($dir = $VC->queryDir($where));

        Chora::checkError($dir->browseDir());

        $dir->applySort();

        Chora::checkError($dirList =& $dir->queryDirList());
        Chora::checkError($fileList = $dir->queryFileList());

        if($recurse)
        {
            foreach ($dirList as $currentDir)
            {
                if($where == '')
                {
                    $next_dir = "$currentDir";
                }
                else 
                {
                    $next_dir = "$where/$currentDir";
                }

                $zipfile->add_dir("$repository_name/$next_dir");

                CVSLib_Checkout::zip_dir($next_dir, &$zipfile, $recurse);

                if ($conf['hide_restricted'] && Chora::isRestricted($currentDir)) 
                {
                    continue;
                }
            }
        }

        foreach ($fileList as $currFile) 
        {
            if ($conf['hide_restricted'] && Chora::isRestricted($currFile->queryName()))
            {
                continue;
            }

            $lg = $currFile->queryLastLog();
            if (PEAR::isError($lg)) 
            {
                continue;
            }

            $filename = $currFile->queryName();

            $latest_rev = $currFile->queryHead();
    
            $base_dir = $VC->sourceroot();             //jcs  
        
            if($where == '') $fullname = $base_dir.'/'.$filename;
            else $fullname = $base_dir.'/'.$where.'/'.$filename;
        
            // Create the CVSLib_File object and populate it....

            $file = &CVSLib_File::getFileObject($VC, $fullname, false);

            // Is this a valid revision being requested? 

            if (!VC_Revision::valid($latest_rev)) 
            {
                fatal('404 Not Found', "Revision Not Found: $r is not a valid RCS revision number");
            }

            $file_ptr = CVSLib_Checkout::get($fullname.',v', $latest_rev);

            if (is_object($file_ptr))//jcs && $file_ptr->id() == CVSLIB_ERROR) 
            {
                Chora::checkError($file_ptr);
            }

            // Extract contents...

            $content = '';
        
            while ($line = fgets($file_ptr)) 
            {
                $content .= $line;
            }
        
            @fclose($file_ptr);

            // Zip contents...

            if($where == '')
            {
                $zipfile->add_file($content, "$repository_name/$filename");
                $zip_log = $zip_log . "$latest_rev\t$filename\r\n";
            }
            else
            {
                $zipfile->add_file($content, "$repository_name/$where/$filename");
                $zip_log = $zip_log . "$latest_rev\t$where/$filename\r\n";
            }

        }

        $recurse_count--;

        if($recurse_count == 0)
        {
            $zipfile->add_file($zip_log, "Versions.log");
        }
    }


}

/**
 * CVSLib diff class.
 *
 * See the README file that came with this library for more
 * information, and read the inline documentation.
 *
 * Copyright Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_Diff {

    /**
     * Obtain the differences between two revisions within a file.
     *
     * @param file CVSLib_File object for the desired file
     * @param rev1 Original revision number to compare from
     * @param rev2 New revision number to compare against
     * @param type Constant which indicates the type of diff (e.g. unified)
     * @param num  Number of lines to be used in context and unified diff
     * @param ws   Show whitespace in the diff?
     *
     * @return false on failure, or a string containing the diff on success
     */
    function get(&$file, $rev1, $rev2, $type = SVNLIB_DIFF_CONTEXT, $num = 3, $ws = true)
    {
        /* Make sure that the file parameter is valid */
        if (is_a($file, 'PEAR_Error')) {
            return false;
        }

        /* Check that the revision numbers are valid */
        $rev1 = VC_Revision::valid($rev1)?$rev1:'1.1';
        $rev2 = VC_Revision::valid($rev1)?$rev2:'1.1';

        $fullName = $file->queryFullPath();
        $diff = array();
        $options = '-kk ';
        if (!$ws) {
            $opts = ' -bB ';
            $options .= $opts;
        } else {
            $opts = '';
        }
        switch ($type) {
        case 'context':
            $options = $opts . '-p --context=' . $num;
            break;

        case 'unified':
            $options = $opts . '-p --unified=' . $num;
            break;

        case 'column':
            $options = $opts . '--side-by-side --width=120';
            break;

        case 'ed':
            $options = $opts . '-e';
            break;
        }

        // TODO: add options for $hr options - however these may not
        // be compatible with some diffs - avsm
        $command = $GLOBALS['conf']['paths']['rcsdiff'] . " $options -r$rev1 -r$rev2 \"" . escapeshellcmd($fullName) . '" 2>&1';
        if (OS_WINDOWS) {
            $command .= ' < "' . __FILE__ . '"';
        }

        exec($command, $diff, $retval);
        return ($retval > 0) ? $diff : array();
    }

}

/**
 * CVSLib directory class.
 *
 * See the README file that came with this library for more
 * information, and read the inline documentation.
 *
 * Copyright Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_Directory {

    var $rep;
    var $dirName;
    var $files;
    var $atticFiles;
    var $mergedFiles;
    var $dirs;
    var $parent;
    var $moduleName;

    /**
     * Create a CVS Directory object to store information
     * about the files in a single directory in the repository
     *
     * @param string                   $dn  Path to the directory.
     * @param object CVSLib_Directory  $pn  (optional) The parent CVSLib_Directory to this one.
     */
    function CVSLib_Directory($rep, $dn, $pn = '')
    {
        $this->rep = &$rep;
        $this->parent = &$pn;
        $this->moduleName = $dn;
        $this->dirName = $rep->sourceroot() . "/$dn";
        $this->files = array();
        $this->dirs = array();
    }

    /**
     * Return fully qualified pathname to this directory with no
     * trailing /.
     *
     * @return Pathname of this directory
     */
    function queryDir()
    {
        return $this->dirName;
    }

    function &queryDirList()
    {
        reset($this->dirs);
        return $this->dirs;
    }

    function &queryFileList($showattic = false)
    {
        if ($showattic && isset($this->mergedFiles)) {
            return $this->mergedFiles;
        } else {
            return $this->files;
        }
    }

    /**
     * Tell the object to open and browse its current directory, and
     * retrieve a list of all the objects in there.  It then populates
     * the file/directory stack and makes it available for retrieval.
     *
     * @return PEAR_Error object on an error, 1 on success.
     */
    function browseDir($quicklog = true, $showattic = false)
    {
        /* Make sure we are trying to list a directory */
        if (!@is_dir($this->dirName)) {
            return PEAR::raiseError(_("Unable to find directory"));
        }

        /* Open the directory for reading its contents */
        if (!($DIR = @opendir($this->dirName))) {
            global $where;
            $errmsg = (!empty($php_errormsg)) ? $php_errormsg : _("Permission Denied");
            return PEAR::raiseError("$where: $errmsg");
        }

        /* Skip over . and .. directories, they are always first. */
        readdir($DIR);
        readdir($DIR);

        /* Create two arrays - one of all the files, and the other of
         * all the dirs */
        while (($name = readdir($DIR)) !== false) {
            $path = $this->dirName . '/' . $name;

            if (@is_dir($path)) {
                /* Skip Attic directory */
                if ($name != 'Attic') {
                    $this->dirs[] = $name;
                }
            } else if (@is_file($path) && substr($name, -2) == ',v') {
                /* Spawn a new file object to represent this file */
                $fl = &$this->rep->getFileObject(substr($path, strlen($this->rep->sourceroot()), -2), $quicklog);
                if (!is_a($fl, 'PEAR_Error')) {
                    $this->files[] = $fl;
                }
            }
        }

        /* Close the filehandle; we've now got a list of dirs and files */
        closedir($DIR);

        /* If we want to merge the attic, add it in here */
        if ($showattic) {
            $atticDir = new CVSLib_Directory($this->rep, $this->moduleName . '/Attic', $this);
            if ($atticDir->browseDir($quicklog) == 1) {
                $this->atticFiles = &$atticDir->queryFileList();
                $this->mergedFiles = array_merge($this->files, $this->atticFiles);
            }
        }

        return true;
    }

    /**
     * Sort the contents of the directory in a given fashion and
     * order.
     *
     * @param $how Of the form CVSLIB_SORT_* where * can be:
     *             NONE, NAME, AGE, REV for sorting by name, age or revision.
     * @param $dir Of the form CVSLIB_SORT_* where * can be:
     *             ASCENDING, DESCENDING for the order of the sort.
     */
    function applySort($how = CVSLIB_SORT_NONE, $dir = CVSLIB_SORT_ASCENDING)
    {
        /* TODO: this code looks very inefficient! optimise... - avsm */
        // assume by name for the moment
        natcasesort($this->dirs);
        $this->doFileSort($this->files, $how, $dir);
        if (isset($this->atticFiles)) {
            $this->doFileSort($this->atticFiles, $how, $dir);
        }
        if (isset($this->mergedFiles)) {
            $this->doFileSort($this->mergedFiles, $how, $dir);
        }
        if ($dir == CVSLIB_SORT_DESCENDING) {
            $this->dirs = array_reverse($this->dirs);
            $this->files = array_reverse($this->files);
            if (isset($this->mergedFiles)) {
                $this->mergedFiles = array_reverse($this->mergedFiles);
            }
        }
    }

    function doFileSort(&$fileList, $how = CVSLIB_SORT_NONE, $dir = CVSLIB_SORT_ASCENDING)
    {
        switch ($how) {
        case CVSLIB_SORT_AGE:
            usort($fileList, array($this, 'fileAgeSort'));
            break;

        case CVSLIB_SORT_NAME:
            usort($fileList, array($this, 'fileNameSort'));
            break;

        case CVSLIB_SORT_AUTHOR:
            usort($fileList, array($this, 'fileAuthorSort'));
            break;

        case CVSLIB_SORT_REV:
            usort($fileList, array($this, 'fileRevSort'));
            break;

        case CVSLIB_SORT_NONE:
        default:
            break;
        }
    }

    /**
     * Sort function for ascending age.
     */
    function fileAgeSort($a, $b)
    {
        $aa = $a->queryLastLog();
        $bb = $b->queryLastLog();
        if ($aa->queryDate() == $bb->queryDate()) {
            return 0;
        } else {
            return ($aa->queryDate() < $bb->queryDate()) ? 1 : -1;
        }
    }

    /**
     * Sort function by author name.
     */
    function fileAuthorSort($a, $b)
    {
        $aa = $a->queryLastLog();
        $bb = $b->queryLastLog();
        if ($aa->queryAuthor() == $bb->queryAuthor()) {
            return 0;
        } else {
            return ($aa->queryAuthor() > $bb->queryAuthor()) ? 1 : -1;
        }
    }

    /**
     * Sort function for ascending filename.
     */
    function fileNameSort($a, $b)
    {
        return strcasecmp($a->name, $b->name);
    }

    /**
     * Sort function for ascending revision.
     */
    function fileRevSort($a, $b)
    {
        return VC_Revision::cmp($a->queryHead(), $b->queryHead());
    }

}

/**
 * CVSLib_file class.
 *
 * See the README file that came with this library for more
 * information, and read the inline documentation.
 *
 * Copyright Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_File extends VC_File {

    var $rep;
    var $dir;
    var $name;
    var $logs;
    var $revs;
    var $head;
    var $quicklog;
    var $symrev;
    var $revsym;
    var $branches;

    /**
     * Create a repository file object, and give it information about
     * what its parent directory and repository objects are.
     *
     * @param string $fl  Full path to this file.
     */
    function CVSLib_File($fl, $quicklog = false)
    {
        $fl .= ',v';
        $this->rep = &$GLOBALS['VC'];
        $this->name = basename($fl);
        $this->dir = dirname($fl);
        $this->logs = array();
        $this->quicklog = $quicklog;
        $this->revs = array();
        $this->branches = array();
    }

    function &getFileObject(&$rep, $filename, $quicklog = false)
    {
        /**
         * The version of the cached data. Increment this whenever the
         * internal storage format changes, such that we must
         * invalidate prior cached data.
         *
         * @var integer $_cacheVersion
         */
        $_cacheVersion = 1;

        static $cache;

        if (is_null($cache)) {
            global $conf;
            require_once HORDE_LIBS . 'Horde/Cache.php';
            $cache = &Horde_Cache::singleton($conf['cache']['driver'], Horde::getDriverConfig('cache', $conf['cache']['driver']));
        }
        $cacheId = $filename . '_f' . (int)$quicklog . '_v' . $_cacheVersion;
        $fileOb = unserialize($cache->getData($cacheId, "serialize(CVSLib_File::_getFileObject('$filename', '$quicklog'))",
                                              time() - @filemtime($filename . ',v')));
        $fileOb->rep = &$rep;
        return $fileOb;
        
    }

    function &_getFileObject($filename, $quicklog = false)
    {
        $fileOb = &new CVSLib_File($filename, $quicklog);
        Chora::checkError($fileOb->getBrowseInfo());
        $fileOb->applySort(CVSLIB_SORT_AGE);
        return $fileOb;
    }

    /**
     * If this file is present in an Attic directory, this indicates
     * it.
     *
     * @return true if file is in the Attic, and false otherwise
     */
    function isAtticFile()
    {
        return substr($this->dir, -5) == 'Attic';
    }

    /**
     * Returns the name of the current file as in the repository
     *
     * @return Filename (without the path)
     */
    function queryRepositoryName()
    {
        return $this->name;
    }

    /**
     * Returns name of the current file without the repository
     * extensions (usually ,v)
     *
     * @return Filename without repository extension
     */
    function queryName()
    {
        return preg_replace('/,v$/', '', $this->name);
    }

    /**
     * Return the last revision of the current file on the HEAD branch
     *
     * @return Last revision of the current file
     */
    function queryRevision()
    {
        if (!isset($this->revs[0])) {
            return PEAR::raiseError(_("No revisions"));
        }
        return $this->revs[0];
    }

    function queryPreviousRevision($rev)
    {
        return VC_Revision::prev($rev);
    }

    /**
     * Return the HEAD (most recent) revision number for this file.
     *
     * @return HEAD revision number
     */
    function queryHead()
    {
        return $this->head;
    }

    /**
     * Return the last VC_log object in the file.
     *
     * @return VC_log of the last entry in the file
     */
    function queryLastLog()
    {
        if (!isset($this->revs[0]) || !isset($this->logs[$this->revs[0]])) {
            return PEAR::raiseError(_("No revisions"));
        }
        return $this->logs[$this->revs[0]];
    }

    /**
     * Sort the list of VC_log objects that this file contains.
     *
     * @param how CVSLIB_SORT_REV (sort by revision),
     *            CVSLIB_SORT_NAME (sort by author name),
     *            CVSLIB_SORT_AGE (sort by commit date)
     */
    function applySort($how = CVSLIB_SORT_REV)
    {
        switch ($how) {
        case CVSLIB_SORT_REV:
            $func = 'Revision';
            break;

        case CVSLIB_SORT_NAME:
            $func = 'Name';
            break;

        case CVSLIB_SORT_AGE:
            $func = 'Age';
            break;

        default:
            $func = 'Revision';
        }
        uasort($this->logs, array($this, "sortBy$func"));
        return true;
    }

    /**
     * The sortBy*() functions are internally used by applySort.
     */
    function sortByRevision($a, $b)
    {
        return VC_Revision::cmp($b->rev, $a->rev);
    }

    function sortByAge($a, $b)
    {
        return $b->date - $a->date;
    }

    function sortByName($a, $b)
    {
        return strcmp($a->author, $b->author);
    }

    /**
     * Populate the object with information about the revisions logs
     * and dates of the file.
     *
     * @return PEAR_Error object on error, or true on success
     */
    function getBrowseInfo()
    {
        /* Check that we are actually in the filesystem */
        if (!is_file($this->queryFullPath())) {
            return PEAR::raiseError(_("File Not Found"));
        }

        /* Call the RCS rlog command to retrieve the file information */
        $flag = $this->quicklog ? ' -r ' : ' ';
        $Q = OS_WINDOWS ? '"' : "'" ;
        
        $cmd = $GLOBALS['conf']['paths']['rlog'] . $flag . $Q . $this->queryFullPath() . $Q;

        exec($cmd, $return_array, $retval);
        if ($retval) {
            return PEAR::raiseError(_("Failed to spawn rlog to retrieve file log information"));
        }

        $accum = array();
        $symrev = array();
        $revsym = array();
        $state = 'init';
        $skip_next = false;
        
        foreach ($return_array as $line) {

            switch ($state) {
            case 'init':

                if (!strncmp('head: ', $line, 6)) {
                    $this->head = substr($line, 6);
                } else if (!strncmp('branch:', $line, 7)) {
                    $state = 'rev';
                }
                break;

            case 'rev':
                // Remove any locking information
                if($skip_next){
                    $skip_next = false;
                    continue;
                }
                if(preg_match("/locks:(\s+)(\w+)(\s*)/", $line)){
                    $line = "locks;";
                    $skip_next = true;
                }
                if (!strncmp('----------', $line, 10)) {
                    $state = 'info';
                    $this->symrev = $symrev;
                    $this->revsym = $revsym;
                } else if (preg_match("/^\s+([^:]+):\s+([\d\.]+)/", $line ,$regs)) {
                    // Check to see if this is a branch
                    if (preg_match('/^(\d+(\.\d+)+)\.0\.(\d+)$/',$regs[2])) {
                        $branchRev = VC_Revision::toBranch($regs[2]);
                        if (!isset($this->branches[$branchRev])) {
                            $this->branches[$branchRev] = $regs[1];
                        }
                    } else {
                        $symrev[$regs[1]] = $regs[2];
                        if (empty($revsym[$regs[2]])) $revsym[$regs[2]]=array();
                        array_push($revsym[$regs[2]], $regs[1]);
                    }
                }
                break;

            case 'info':
                // Remove any locked by information
                $line = preg_replace("/(\s+)locked by: (\w+);/", "", $line);
                if (strncmp('==============================', $line, 30) &&
                    strcmp('----------------------------', $line)) {
                    $accum[] = $line;
                } else if (count($accum)) {
                    // spawn a new VC_log object and add it to the logs hash
                    $log = &new CVSLib_Log($this);
                    $err = $log->processLog($accum);
                    // TODO: error checks - avsm
                    $this->logs[$log->queryRevision()] = $log;
                    array_push($this->revs, $log->queryRevision());
                    $accum = array();
                }
                break;
            }
        }

        return true;
    }

    /**
     * Return the fully qualified filename of this object.
     *
     * @return Fully qualified filename of this object
     */
    function queryFullPath()
    {
        return $this->dir . DIRECTORY_SEPARATOR . $this->name;
    }

    /**
     * Return the name of this file relative to its sourceroot.
     *
     * @return string  Pathname relative to the sourceroot.
     */
    function queryModulePath()
    {
        return preg_replace('|^'. $this->rep->sourceroot() . '/?(.*),v$|', '\1', $this->queryFullPath());
    }

}

/**
 * CVSLib log class.
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_Log {

    var $rep, $file, $tags, $rev, $date, $log, $author, $state, $lines, $branches;

    /**
     *
     */
    function CVSLib_Log(&$fl)
    {
        $this->file = &$fl;
        $this->branches = array();
    }

    function processLog($raw)
    {
        /* Initialise a simple state machine to parse the output of rlog */
        $state = 'init';
        while (!empty($raw) && $state != 'done') {
            switch ($state) {
            /* Found filename, now looking for the revision number */
            case 'init':
                $line = array_shift($raw);
                if (preg_match("/revision (.+)$/", $line, $parts)) {
                    $this->rev = $parts[1];
                    $state = 'date';
                }
                break;

            /* Found revision and filename, now looking for date */
            case 'date':
                
                $line = array_shift($raw);
                if (preg_match("|^date:\s+(\d+)[-/](\d+)[-/](\d+)\s+(\d+):(\d+):(\d+).*?;\s+author:\s+(\S+);\s+state:\s+(\S+);(\s+lines:\s+([0-9\s+-]+))?|", $line, $parts)) {
                    $this->date = gmmktime($parts[4], $parts[5], $parts[6], $parts[2], $parts[3], $parts[1]);
                    $this->author = $parts[7];
                    $this->state = $parts[8];
                    $this->lines = isset($parts[10]) ? $parts[10] : '';
                    $state = 'branches';
                }
                break;

            /* Look for a branch point here - format is 'branches:  x.y.z;  a.b.c;' */
            case 'branches':
                /* If we find a branch tag, process and pop it,
                   otherwise leave input stream untouched */
                if (!empty($raw) && preg_match("/^branches:\s+(.*)/", $raw[0], $br)) {
                    /* Get the list of branches from the string, and
                     * push valid revisions into the branches array */
                    $brs = preg_split('/;\s*/', $br[1]);
                    foreach ($brs as $brpoint) {
                        if (VC_Revision::valid($brpoint)) {
                            $this->branches[] = $brpoint;
                        }
                    }
                    array_shift($raw);

                }

                $state = 'done';
                break;

            default:
            }
        }

        /* Assume the rest of the lines are the log message */
        $this->log = implode("\n", $raw);
        $this->tags = @$this->file->revsym[$this->rev];
        if (empty($this->tags)) {
            $this->tags = array();
        }
    }

    function queryDate()
    {
        return $this->date;
    }

    function queryRevision()
    {
        return $this->rev;
    }

    function queryAuthor()
    {
        return $this->author;
    }

    function queryLog()
    {
        return $this->log;
    }

    function queryChangedLines()
    {
        return isset($this->lines) ? ($this->lines) : '';
    }

    /**
     * Given a branch revision number, this function remaps it
     * accordingly, and performs a lookup on the file object to
     * return the symbolic name(s) of that branch in the tree.
     *
     * @return hash of symbolic names => branch numbers
     */
    function querySymbolicBranches()
    {
        $symBranches = array();
        foreach ($this->branches as $branch) {
            $parts = explode('.', $branch);
            $last = array_pop($parts);
            $parts[] = '0';
            $parts[] = $last;
            $rev = implode('.', $parts);
            if (isset($this->file->branches[$branch])) {
                $symBranches[$this->file->branches[$branch]] = $branch;
            }
        }
        return $symBranches;
    }

}

/**
 * CVSLib_Patchset class.
 *
 * Copyright Anil Madhavapeddy, <anil@recoil.org>
 *
 * @author  Anil Madhavapeddy <anil@recoil.org>
 * @since   Chora 0.1
 * @package VC
 */
class CVSLib_Patchset {

    var $_dir;
    var $_name;
    var $_patchsets = array();

    /**
     * Create a patchset object.
     *
     * @param string $file  The filename to get patchsets for.
     */
    function CVSLib_Patchset($file)
    {
        $this->_name = basename($file);
        $this->_dir = dirname($file);
    }

    function &getPatchsetObject(&$rep, $filename)
    {
        /**
         * The version of the cached data. Increment this whenever the
         * internal storage format changes, such that we must
         * invalidate prior cached data.
         *
         * @var integer $_cacheVersion
         */
        $_cacheVersion = 1;

        static $cache;

        if (is_null($cache)) {
            global $conf;
            require_once HORDE_LIBS . 'Horde/Cache.php';
            $cache = &Horde_Cache::singleton($conf['cache']['driver'], Horde::getDriverConfig('cache', $conf['cache']['driver']));
        }

        $cacheId = $filename . '_f' . '_v' . $_cacheVersion;
        $psOb = unserialize($cache->getData($cacheId, "serialize(CVSLib_Patchset::_getPatchsetObject('$filename', '" . $rep->sourceroot() . "'))",
                                            time() - @filemtime($filename . ',v')));
        return $psOb;
    }

    function &_getPatchsetObject($filename, $repository)
    {
        $psOb = &new CVSLib_Patchset($filename);
        Chora::checkError($psOb->getPatchsets($repository));
        return $psOb;
    }

    /**
     * Populate the object with information about the patchsets that
     * this file is involved in.
     *
     * @param string $repository  The full repository location.
     *
     * @return mixed  PEAR_Error object on error, or true on success.
     */
    function getPatchsets($repository)
    {
        global $conf;

        /* Check that we are actually in the filesystem. */
        if (!is_file($this->getFullPath() . ',v')) {
            return PEAR::raiseError(_("File Not Found"));
        }

        /* Call cvsps to retrieve all patchsets for this file. */
        $Q = OS_WINDOWS ? '"' : "'";

        $HOME = !empty($conf['paths']['cvsps_home']) ?
            'HOME=' . $conf['paths']['cvsps_home'] . ' ' :
            '';

        $cmd = $HOME . $GLOBALS['conf']['paths']['cvsps'] . ' -u --cvs-direct --root ' . $Q . $repository . $Q . ' -f ' . $Q . $this->_name . $Q . ' ' . $Q . $this->_dir . $Q;
        exec($cmd, $return_array, $retval);
        if ($retval) {
            return PEAR::raiseError(_("Failed to spawn cvsps to retrieve patchset information"));
        }

        $this->_patchsets = array();
        $state = 'begin';
        foreach ($return_array as $line) {
            $line = trim($line);
            if ($line == '---------------------') {
                $state = 'begin';
                continue;
            }

            switch ($state) {
            case 'begin':
                $id = str_replace('PatchSet ', '', $line);
                $this->_patchsets[$id] = array();
                $state = 'info';
                break;

            case 'info':
                $info = explode(':', $line, 2);
                switch ($info[0]) {
                case 'Date':
                    $this->_patchsets[$id]['date'] = trim($info[1]);
                    break;

                case 'Author':
                    $this->_patchsets[$id]['author'] = trim($info[1]);
                    break;

                case 'Branch':
                    $this->_patchsets[$id]['branch'] = trim($info[1]);
                    break;

                case 'Tag':
                    $this->_patchsets[$id]['tag'] = trim($info[1]);
                    break;

                case 'Log':
                    $state = 'log';
                    $this->_patchsets[$id]['log'] = '';
                    break;
                }
                break;

            case 'log':
                if ($line == 'Members:') {
                    $state = 'members';
                    $this->_patchsets[$id]['log'] = trim($this->_patchsets[$id]['log']);
                    $this->_patchsets[$id]['members'] = array();
                } else {
                    $this->_patchsets[$id]['log'] .= $line . "\n";
                }
                break;

            case 'members':
                if (!empty($line)) {
                    $parts = explode(':', $line);
                    $revs = explode('->', $parts[1]);
                    $this->_patchsets[$id]['members'][] = array('file' => $parts[0],
                                                                'from' => $revs[0],
                                                                'to' => $revs[1]);
                }
                break;
            }
        }

        return true;
    }

    /**
     * Return the fully qualified filename of this object.
     *
     * @return Fully qualified filename of this object
     */
    function getFullPath()
    {
        return $this->_dir . DIRECTORY_SEPARATOR . $this->_name;
    }

}
